/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
   MODULO LDAP                                                                   
   Programado por Lucas Priori                                                   
   Versao: 1.00                                                                  
                                                                                 
  Linux:                                                                         
  32 Bits: gcc -shared -nostartfiles selor_ldap.c -o selor_ldap.so -lldap -Wall  
  64 Bits: gcc -shared -nostartfiles selor_ldap.c -o selor_ldap.so -DA64 -lldap -Wall -fPIC
                                                                                 
  FreeBSD:                                                                       
  32 Bits: gcc -shared -nostartfiles selor_ldap.c -o selor_ldap.so -DFreeBSD -lldap -Wall
  64 Bits: gcc -shared -nostartfiles selor_ldap.c -o selor_ldap.so -DFreeBSD -DA64 -lldap -Wall -fPIC
                                                                                 
  ****************************************************************************** 
  Este programa  um modulo LDAP para o MTA SELOR                                
  Copyright (C) 2012,2013,2014  Lucas Priori                                     
                                                                                 
  This program is free software; you can redistribute it and/or                  
  modify it under the terms of the GNU General Public License                    
  as published by the Free Software Foundation; either version 2                 
  of the License, or (at your option) any later version.                         
                                                                                 
  This program is distributed in the hope that it will be useful,                
  but WITHOUT ANY WARRANTY; without even the implied warranty of                 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  
  GNU General Public License for more details.                                   
                                                                                 
  You should have received a copy of the GNU General Public License              
  along with this program; if not, write to the Free Software                    
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<fcntl.h>
#include<unistd.h>
#include<sys/ipc.h>
#include<sys/sem.h>
#include<ldap.h>
#include<stdint.h>
#ifdef FreeBSD
 #include<rpcsvc/crypt.h>
#else
 #include<crypt.h>
#endif

#include "md5.c"

#define MOD_NAME "ldap\0"
#define VERSAO "1.0"
#define TAM_buffer 2048
#define TAM_msg_error 200
#define TAM_user 250
#define TAM_pass 250
#define TAM_home 500
#define TAM_dlocal 500
#define TAM_maildir 100
#define MOD_user_user 0
#define MOD_user_pass 1
#define MOD_user_home 2
#define MOD_user_dlocal 3
#define MAX_cnx 30

//---------------------------------------------------
//      Variaveis GLOBAIS                            
//---------------------------------------------------
// -- Dados do arquivo de configuracao --
char mod_name[50]=MOD_NAME;
char host[100]={0};
char login_dn[400]={0};
char pass[100]={0};
char attr_domain[50]={0};
char attr_mail[50]={0};
char attr_pass[50]={0};
char attr_home[50]={0};
char attr_maildir[50]={0};
char attr_quota[50]={0};
char attr_dlocal[50]={0};
char attr_dstalias[50]={0};
char attr_dstadomain[50]={0};
char sch_domain_base[250]={0};
char sch_domain_filter[250]={0};
char sch_adomain_base[250]={0};
char sch_adomain_filter[250]={0};
char sch_user_base[250]={0};
char sch_user_filter[250]={0};
char sch_alias_base[250]={0};
char sch_alias_filter[250]={0};
char *attribs_domain[]={attr_domain, NULL};
char *attribs_adomain[]={attr_dstadomain, NULL};
char *attribs_user[]={attr_mail, attr_pass, attr_home, attr_maildir, attr_quota, NULL, NULL};
char *attribs_alias[]={attr_dstalias, NULL};

short sem_num=0;
short atual_cnx=0;               // Variavel que armazena a atual conexao ldap

int num_cnx=5;
int tam_attr_domain=0;
int tam_attr_mail=0;
int tam_attr_pass=0;
int tam_attr_home=0;
int tam_attr_maildir=0;
int tam_attr_quota=0;
int tam_attr_dlocal=0;
int tam_attr_dstalias=0;
int tam_attr_dstadomain=0;

// --
// --- Semaforo ---
key_t chave;
union semun {
    int val;
    struct semid_ds *buf;
    unsigned short  *array;
}sm_ctl;
struct sembuf sm_travar,sm_liberar;
int semid;
// --
LDAP *ld[MAX_cnx+1];
LDAPMessage *res, *msg;
BerElement *ber;
char tab_base64_reversa[256]={ 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 62,  0,  0,  0, 63,
                              52, 53, 54, 55, 56, 57, 58, 59, 60, 61,  0,  0,  0,  0,  0,  0,
                               0,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
                              15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,  0,  0,  0,  0,  0,
                               0, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
                              41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
                               0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};
struct berval **str_berval;
int ldap_proto_versao=LDAP_VERSION3;
char buffer[TAM_buffer + 1];
char linha[TAM_buffer + 1];
char msg_error[TAM_msg_error + 1];
char *str_dn;
char *attr;
int msg_type;
// --- CACHE DE USUARIO ---
char cache_user[TAM_user]={0};
char cache_pass[TAM_pass]={0};
char cache_home[TAM_home]={0};
char cache_dlocal[TAM_dlocal]={0};
char cache_maildir[TAM_maildir]={0};
unsigned long long int cache_quota=0;
// ---
char ret=0;
char linha_quota[100];
int travador=0;
int arq_quota=0;
int tam_user=0;
//int num_alias=0;
int b_lidos=0;
unsigned int cont1=0;
unsigned int cont2=0;
unsigned int cont3=0;
unsigned int num_registros=0;
long long int tam=0;
long long int num=0;

// =========================
//  Funcoes uteis do modulo 
// =========================

//---------------------------------------------------
//Funcao para comparar duas strings desconciderando  
//a diferena entre maiusculas e minusculas          
//Retornos:                                          
//0 = strings so iguais                             
//1 = strings so diferentes                         
int cmp_str(char *string1, char *string2, int num_bytes){
 if(!num_bytes){
  if(*string1==*string2)return 0;
  return 1;
 }
 int ret_val=0;
 
 #ifdef A64
  asm(
   "mov  %0,%%rdi           \n"
   "mov  %1,%%rsi           \n"
  "CMP_STR_proxbyte:        \n"
   "movb (%%rsi),%%al       \n"
   "cmpb (%%rdi),%%al       \n"
   "jne  CMP_STR_trocar     \n"
  "CMP_STR_incr:            \n"
   "inc  %%rsi              \n"
   "inc  %%rdi              \n"
   "loop CMP_STR_proxbyte   \n"
   "jmp  CMP_STR_term       \n"
  "CMP_STR_trocar:          \n"
   "xorb $32,%%al           \n"
   "cmpb (%%rdi),%%al       \n"
   "je   CMP_STR_incr       \n"
  "CMP_STR_term:            \n"
   "testw %%cx,%%cx         \n"
   "jz   CMP_STR_fim        \n"
   "movw $1,%%cx            \n"
  "CMP_STR_fim:             \n"
   "movw  %%cx,%3           \n"
   :
   :"m"(string1),"m"(string2),"c"(num_bytes),"m"(ret_val)
   :"rdi","rsi","rax"
  );
 #else
  asm(
   "movl %0,%%edi           \n"
   "movl %1,%%esi           \n"
  "CMP_STR_proxbyte:        \n"
   "movb (%%esi),%%al       \n"
   "cmpb (%%edi),%%al       \n"
   "jne  CMP_STR_trocar     \n"
  "CMP_STR_incr:            \n"
   "incl %%esi              \n"
   "incl %%edi              \n"
   "loop CMP_STR_proxbyte   \n"
   "jmp  CMP_STR_term       \n"
  "CMP_STR_trocar:          \n"
   "xorb $32,%%al           \n"
   "cmpb (%%edi),%%al       \n"
   "je   CMP_STR_incr       \n"
  "CMP_STR_term:            \n"
   "testw %%cx,%%cx         \n"
   "jz   CMP_STR_fim        \n"
   "movw $1,%%cx            \n"
  "CMP_STR_fim:             \n"
   "movw  %%cx,%3           \n"
   :
   :"m"(string1),"m"(string2),"c"(num_bytes),"m"(ret_val)
   :"edi","esi","eax"
  );
 #endif
 return ret_val;
}

//---------------------------------------------------
// Funcao semelhante a strncpy, porem, retorna o     
// numero de bytes copiados, pra quando encontra    
// ZERO e tem                                        
// protecao contra buffer overflow                   
unsigned int local_strncpy(char *destino, char *origem, unsigned int tam_str){
 unsigned int ret_val;
 
 #ifdef A64
   asm(
   "mov  %0,%%rdi           \n"
   "mov  %1,%%rsi           \n"
   "mov  %%rsi,%%rbx        \n"
  "STRNCPY_prox:            \n"
   "lodsb                   \n"
   "stosb                   \n"
   "testb $255,%%al         \n"
   "jz    STRNCPY_fim       \n"
   "loop  STRNCPY_prox      \n"
   "inc  %%rsi              \n"
  "STRNCPY_fim:             \n"
   "movb  $0,(%%rdi)        \n"
   "sub  %%rbx,%%rsi        \n"
   "dec  %%rsi              \n"   
   "mov  %%rsi,%3           \n"
   :
   :"m"(destino), "m"(origem), "c"(tam_str), "m"(ret_val)
   :"rdi", "rsi", "ax", "rbx"
  );
 #else
  asm(
   "movl  %0,%%edi          \n"
   "movl  %1,%%esi          \n"
   "movl  %%esi,%%ebx       \n"
  "STRNCPY_prox:            \n"
   "lodsb                   \n"
   "stosb                   \n"
   "testb $255,%%al         \n"
   "jz    STRNCPY_fim       \n"
   "loop  STRNCPY_prox      \n"
   "incl  %%esi             \n"
  "STRNCPY_fim:             \n"
   "movb  $0,(%%edi)        \n"
   "subl  %%ebx,%%esi       \n"
   "decl  %%esi             \n"   
   "movl  %%esi,%3          \n"
   :
   :"m"(destino), "m"(origem), "c"(tam_str), "m"(ret_val)
   :"edi", "esi", "ax" , "ebx"
  );
 #endif
 
 return ret_val;
}


//---------------------------------------------------
//Funo para decodificar MIME64                     
//Retorna o tamanho do novo buffer decodificado      
unsigned int base64_decod(char *retorno,char *str_base64){
 unsigned int  tamanho=0;
 unsigned int  diferenca=0;
 unsigned int  tam_temp;
 char *base=tab_base64_reversa;
 tam_temp=strlen(str_base64);
 if(tam_temp<4)return 0;
 tamanho=tam_temp>>2;
 
 #ifdef A64
  asm(
   "mov  %1,%%rdi               \n"
   "mov  %2,%%rsi               \n"
   "mov  %3,%%rbx               \n"
   "add  $4,%%rsi               \n"
  "DECProx_casa:                \n"
   "push  %%rcx                 \n"
   "mov  $4,%%ecx               \n"
  "DECProx_byte:                \n"
   "dec  %%rsi                  \n"
   "mov  (%%rsi),%%al           \n"
   "xlatb                       \n"
   "ror  $6,%%eax               \n"
   "loop DECProx_byte           \n"
   "xchg  %%ah,%%al             \n"
   "ror  $16,%%eax              \n"
   "xchg  %%ah,%%al             \n"
   "mov  %%eax,(%%rdi)          \n"
   "add  $3,%%rdi               \n"
   "add  $8,%%rsi               \n"
   "pop  %%rcx                  \n"
   "loop DECProx_casa           \n"
   "sub  $5,%%rsi               \n"
   "xor  %%ecx,%%ecx            \n"
   "DECAcha_fim:                \n"
   "cmpb $61,(%%rsi)            \n"
   "jne  DECFim                 \n"
   "dec  %%rsi                  \n"
   "dec  %%rdi                  \n"
   "inc  %%ecx                  \n"
   "jmp  DECAcha_fim            \n"
   "DECFim:                     \n"
   "movb $0,(%%rdi)             \n"
   :"=c"(diferenca)
   :"m"(retorno),"m"(str_base64),"m"(base),"c"(tamanho)
   :"rdi","rsi","rax","rbx"
  );
 #else
  asm(
   "movl %1,%%edi               \n"
   "movl %2,%%esi               \n"
   "movl %3,%%ebx               \n"
   "addl $4,%%esi               \n"
  "DECProx_casa:                \n"
   "pushl %%ecx                 \n"
   "movl $4,%%ecx               \n"
  "DECProx_byte:                \n"
   "decl %%esi                  \n"
   "movb (%%esi),%%al           \n"
   "xlatb                       \n"
   "rorl $6,%%eax               \n"
   "loop DECProx_byte           \n"
   "xchgb %%ah,%%al             \n"
   "rorl $16,%%eax              \n"
   "xchgb %%ah,%%al             \n"
   "movl %%eax,(%%edi)          \n"
   "addl $3,%%edi               \n"
   "addl $8,%%esi               \n"
   "popl %%ecx                  \n"
   "loop DECProx_casa           \n"
   "subl $5,%%esi               \n"
   "xorl %%ecx,%%ecx            \n"
   "DECAcha_fim:                \n"
   "cmpb $61,(%%esi)            \n"
   "jne  DECFim                 \n"
   "decl %%esi                  \n"
   "decl %%edi                  \n"
   "incl %%ecx                  \n"
   "jmp  DECAcha_fim            \n"
   "DECFim:                     \n"
   "movb $0,(%%edi)             \n"
   :"=c"(diferenca)
   :"m"(retorno),"m"(str_base64),"m"(base),"c"(tamanho)
   :"edi","esi","eax","ebx"
  );
 #endif
 return tamanho*3-diferenca;
}


//---------------------------------------------------
// Funcao para copiar uma string a outras convertendo
// as variaveis                                      
// %d = Dominio (Domain)                             
// %u = Usuario (User)                               
void cp_str_var(char *destino, int tam_dest, char *origem, char *ptr_dominio, char *ptr_usr){
 int b_copiados=0;
 
 while(*origem && tam_dest > 1){
  if(*origem == '%'){
   if(*++origem == 'u' && ptr_usr){
    b_copiados=local_strncpy(destino, ptr_usr, tam_dest);  // Usuario
    destino+=b_copiados;
    tam_dest-=b_copiados;
   }else if(*origem == 'd' && ptr_dominio){
    b_copiados=local_strncpy(destino, ptr_dominio, tam_dest);  // Dominio
    destino+=b_copiados;
    tam_dest-=b_copiados;
   }else {
    *destino++=*origem;
    if(!*origem)return;
   }
   origem++;
  }else *destino++=*origem++;
 }
 if(tam_dest < 1)*destino=0;
 else *destino=*origem;
}



//======================================================
// Funcao que pega uma conexao ldap disponivel          
void get_free_cnx(void){
 
 if(atual_cnx == MAX_cnx)return;              // Se atual_cnx = MAX_cnx, significa que ja foi crida
                                              // uma conexao extra para atender a este modulo e nos
                                              // a usaremos at o fim do processo                  
 for(sem_num=0; sem_num<num_cnx; sem_num++){
  sm_travar.sem_num=sem_num;
  if(!(semop(semid, &sm_travar, 1))){
   atual_cnx=sem_num;
   return;
  }
 }
 // Nenhuma conexao disponivel. Vamos criar uma e retorna-la em ld[MAX_cnx]                                       
 if(ldap_initialize(&ld[MAX_cnx], host))return;  // Erro iniciando as funcoes LDA
 ldap_set_option(ld[MAX_cnx], LDAP_OPT_PROTOCOL_VERSION, &ldap_proto_versao);
 ldap_simple_bind_s(ld[MAX_cnx], login_dn, pass);
 atual_cnx=MAX_cnx;
}


//======================================================
// Funcao para liberar uma conexao                      
void free_cnx(void){
 
 if(atual_cnx < MAX_cnx){                   // Liberar apenas um semaforo
  sm_liberar.sem_num=atual_cnx;             // Caso a conexao tenha sido aberta em LDAP[MAX_cnx]
  semop(semid, &sm_liberar, 1);             // vamos usa-la ate o fim do processo correte e sera 
 }                                          // fechada automaticamente com o processo morrer     
}



// Funcoes chamadas pela DLlib. Obrigatorias
void _init();  // Chamada quando o modulo  carregado 
void _fini();  // Chamada quando o modulo  desativado

//======================================================
// Funcao chamada pelo MTA assim que  carregado. Igual 
// a _init, porm, aceita um parametro que pode ser um  
// arquivo de configuracao a ler.  o parametro passado 
// dentro do selor.conf                                 
// Retornos:                                            
// 0 = Sem erros, modulo carregado                      
// 1 = Erro                                             
char mod_init(char *param, char *module_name, char *module_version, char *info, int info_buf_size){
 FILE *arq_conf;
 char *ptr;
 int b_lidos=0;
 
 // Abrindo arquivo de configurao
 if(!param){
  strcpy(msg_error, "You must specify a configuration file");
  return 1;
 }
 
 arq_conf=fopen(param, "rb");
 if(!arq_conf){
  snprintf(msg_error, TAM_msg_error, "File not found [%s]", param);
  return 1;
 }
 
 while(!feof(arq_conf)){
  b_lidos=fread(&buffer, 1, TAM_buffer, arq_conf);
  for(cont1=0; cont1<b_lidos; cont1++){
   if(buffer[cont1] == 10){
    linha[cont2]=0;
    if(!cont2)continue;
    if(linha[0] == '#'){
     cont2=0;
     continue;
    }
    if(!cmp_str(linha, "mod_name=", 9)){
     // -- MOD_NAME --
     ptr=&linha[9];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     strncpy(mod_name, ptr, sizeof(mod_name));
    }else if(!cmp_str(linha, "host=", 5)){
     // -- HOST -- 
     ptr=&linha[5];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(host, ptr, sizeof(host));
    }else if(!cmp_str(linha, "login_dn=", 9)){
     // -- LOGIN_DN -- 
     ptr=&linha[9];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(login_dn, ptr, sizeof(login_dn));
    }else if(!cmp_str(linha, "pass=", 5)){
     // -- PASS -- 
     ptr=&linha[5];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(pass, ptr, sizeof(pass));
    }else if(!cmp_str(linha, "num_cnx=", 8)){
     // -- NUM_CNX --
     ptr=&linha[8];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     num_cnx=atoi(ptr);
     if(num_cnx >= MAX_cnx)num_cnx=MAX_cnx;
    }else if(!cmp_str(linha, "attr_domain=", 12)){
     // -- ATTR_DOMAIN -- 
     ptr=&linha[12];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_domain=local_strncpy(attr_domain, ptr, sizeof(attr_domain))+1;
    }else if(!cmp_str(linha, "attr_mail=", 10)){
     // -- ATTR_MAIL -- 
     ptr=&linha[10];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_mail=local_strncpy(attr_mail, ptr, sizeof(attr_mail))+1;
    }else if(!cmp_str(linha, "attr_pass=", 10)){
     // -- ATTR_PASS -- 
     ptr=&linha[10];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_pass=local_strncpy(attr_pass, ptr, sizeof(attr_pass))+1;
    }else if(!cmp_str(linha, "attr_home=", 10)){
     // -- ATTR_HOME -- 
     ptr=&linha[10];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_home=local_strncpy(attr_home, ptr, sizeof(attr_home))+1;
    }else if(!cmp_str(linha, "attr_maildir=", 13)){
     // -- ATTR_MAILDIR -- 
     ptr=&linha[13];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_maildir=local_strncpy(attr_maildir, ptr, sizeof(attr_maildir))+1;
    }else if(!cmp_str(linha, "attr_quota=", 11)){
     // -- ATTR_QUOTA -- 
     ptr=&linha[11];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_quota=local_strncpy(attr_quota, ptr, sizeof(attr_quota))+1;
     attribs_user[5]=attr_quota;
    }else if(!cmp_str(linha, "attr_dlocal=", 12)){
     // -- ATTR_DLOCAL -- 
     ptr=&linha[12];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_dlocal=local_strncpy(attr_dlocal, ptr, sizeof(attr_dlocal))+1;
    }else if(!cmp_str(linha, "attr_dstalias=", 14)){
     // -- ATTR_DSTALIAS -- 
     ptr=&linha[14];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_dstalias=local_strncpy(attr_dstalias, ptr, sizeof(attr_dstalias))+1;
    }else if(!cmp_str(linha, "attr_dstadomain=", 16)){
     // -- ATTR_DSTADOMAIN -- 
     ptr=&linha[16];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     tam_attr_dstadomain=local_strncpy(attr_dstadomain, ptr, sizeof(attr_dstadomain))+1;
    }else if(!cmp_str(linha, "sch_domain_base=", 16)){
     // -- SCH_DOMAIN_BASE -- 
     ptr=&linha[16];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_domain_base, ptr, sizeof(sch_domain_base));
    }else if(!cmp_str(linha, "sch_domain_filter=", 18)){
     // -- SCH_DOMAIN_FILTER -- 
     ptr=&linha[18];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_domain_filter, ptr, sizeof(sch_domain_filter));
    }else if(!cmp_str(linha, "sch_adomain_base=", 17)){
     // -- SCH_ADOMAIN_BASE -- 
     ptr=&linha[17];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_adomain_base, ptr, sizeof(sch_adomain_base));
    }else if(!cmp_str(linha, "sch_adomain_filter=", 19)){
     // -- SCH_ADOMAIN_FILTER -- 
     ptr=&linha[19];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_adomain_filter, ptr, sizeof(sch_adomain_filter));
    }else if(!cmp_str(linha, "sch_user_base=", 14)){
     // -- SCH_USER_BASE -- 
     ptr=&linha[14];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_user_base, ptr, sizeof(sch_user_base));
    }else if(!cmp_str(linha, "sch_user_filter=", 16)){
     // -- SCH_USER_FILTER -- 
     ptr=&linha[16];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_user_filter, ptr, sizeof(sch_user_filter));
    }else if(!cmp_str(linha, "sch_alias_base=", 15)){
     // -- SCH_ALIAS_BASE -- 
     ptr=&linha[15];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_alias_base, ptr, sizeof(sch_alias_base));
    }else if(!cmp_str(linha, "sch_alias_filter=", 17)){
     // -- SCH_ALIAS_FILTER -- 
     ptr=&linha[17];
     while(*ptr == ' ' || *ptr == 9)ptr++;
     local_strncpy(sch_alias_filter, ptr, sizeof(sch_alias_filter));
    }else {
     snprintf(msg_error, TAM_msg_error, "Invalid option [%s]", linha);
     return 1;
    }
    cont2=0;
   }else if(buffer[cont1] != 13)linha[cont2++]=buffer[cont1];
  }
 }
 
 // -- Multiplas conexoes --
 if(num_cnx){
  chave=ftok(param, 'S');
  if((semid=semget(chave, 0, 0660)) != -1){
   // Semaforo ja existe. Vamos remove-lo para criar novamente
   semctl(semid, 0, IPC_RMID, NULL);
  }
  
  // Abrindo semaforo para controlar as conexoes
  if((semid=semget(chave, num_cnx, 0660 | IPC_CREAT)) == -1){
   Sem_error:
   strcpy(msg_error, "Error creating semaphore");
   return 1;
  }
  // Liberando os semaforos com valores positivos
  sm_ctl.val=1;
  for(cont1=0; cont1<num_cnx; cont1++)if(semctl(semid, cont1, SETVAL, sm_ctl))goto Sem_error;
  
  sm_travar.sem_op=-1;
  sm_travar.sem_flg=IPC_NOWAIT | SEM_UNDO;
  sm_liberar.sem_op=1;
  sm_liberar.sem_flg=IPC_NOWAIT | SEM_UNDO;
  
  // -- Abrindo as conexoes com o Banco de Dados --
  for(cont2=0; cont2<num_cnx; cont2++){
   // Conectando na base LDAP
   if(ldap_initialize(&ld[cont2], host)){
    strcpy(msg_error, "Couldn't initialize LDAP functions");
    return 1;
   }
   ldap_set_option(ld[cont2], LDAP_OPT_PROTOCOL_VERSION, &ldap_proto_versao);
   cont1=ldap_simple_bind_s(ld[cont2], login_dn, pass);
   if(cont1 != LDAP_SUCCESS){
    snprintf(msg_error, TAM_msg_error, "LDAP Error: [%s]", ldap_err2string(cont1));
    return 1;
   }
  }
 }
 
 strcpy(module_name, mod_name);
 strcpy(module_version, VERSAO);
 return 0;
}

//======================================================
// Checa se o dominio  local ou nao                    
// Retornos:                                            
// 0 = Nao  local                                      
// 1 = Dominio  local                                  
char mod_dom_local(char *dominio, char *ret_dom){
 
 // Apenas para testes
 //printf("LDAP: mod_dom_local(%s)\n", dominio);
 
 if(sch_domain_base[0]){
  cp_str_var(linha, TAM_buffer, sch_domain_filter, dominio, NULL);
  cp_str_var(buffer, TAM_buffer, sch_adomain_base, dominio, NULL);
  get_free_cnx();
  cont1=ldap_search_ext_s(ld[atual_cnx], buffer, LDAP_SCOPE_ONELEVEL, linha, attribs_domain, 0 , NULL, NULL, NULL, LDAP_NO_LIMIT, &res);
  if(cont1 != LDAP_SUCCESS){
   free_cnx();
   snprintf(msg_error, TAM_msg_error, "LDAP Error: [%s]", ldap_err2string(cont1));
   return -1;
  }
  msg=ldap_first_message(ld[atual_cnx], res);
  if(msg){
   msg_type=ldap_msgtype(msg);
   if(msg_type == LDAP_RES_SEARCH_ENTRY){
    // Lendo atributos
    attr=ldap_first_attribute(ld[atual_cnx], msg, &ber);
    if(attr){
     ldap_memfree(attr);
     ldap_msgfree(msg);
     *ret_dom=0;          // Para indicar que nao  um dominio alias (OBRIGATORIO)
     free_cnx();
     return 1;            // --> Dominio LOCAL      
    }
   } 
  }
  ldap_msgfree(msg);
  free_cnx();
 }
 
 // Vamos procurar um Dominio Alias
 if(sch_adomain_base[0]){
  cp_str_var(linha, TAM_buffer, sch_adomain_filter, dominio, NULL);
  cp_str_var(buffer, TAM_buffer, sch_adomain_base, dominio, NULL);
  get_free_cnx();
  cont1=ldap_search_ext_s(ld[atual_cnx], buffer, LDAP_SCOPE_ONELEVEL, linha, attribs_adomain, 0 , NULL, NULL, NULL, LDAP_NO_LIMIT, &res);
  if(cont1 != LDAP_SUCCESS){
   snprintf(msg_error, TAM_msg_error, "LDAP Error: [%s]", ldap_err2string(cont1));
   free_cnx();
   return -1;
  }
  msg=ldap_first_message(ld[atual_cnx], res);
  if(msg){
   msg_type=ldap_msgtype(msg);
   if(msg_type == LDAP_RES_SEARCH_ENTRY){
    // Lendo atributos
    attr=ldap_first_attribute(ld[atual_cnx], msg, &ber);
    if(attr){
     if((str_berval=ldap_get_values_len(ld[atual_cnx], msg, attr)) != NULL){
      if(str_berval[0])strcpy(ret_dom, str_berval[0]->bv_val);
      ldap_value_free_len(str_berval);
     }
     ldap_memfree(attr);
     ldap_msgfree(msg);
     free_cnx();
     return 1;            // --> Dominio LOCAL      
    }
   } 
  }
  ldap_msgfree(msg);
  free_cnx();
 }
 
 return 0;
}

//======================================================
// Checa se o email  local ou nao                      
// Retornos:                                            
// 0 = Email nao existe                                 
// 1 = Email exite                                      
// 2 = Email existe e forcar uma busca na tab alias     
// -1 = Erro                                            
char mod_user_local(char *email, char *info[], int num_campos){
 char dominio[200];
 char *ptr_aux1;
 
 // Apenas para testes
 //printf("LDAP: mod_user_local(%s)\n", email);
 if(sch_user_base[0]){
  if(!cmp_str(email, cache_user, tam_user+1)){
   // Email no cache
   info[MOD_user_user]=cache_user;
   info[MOD_user_pass]=cache_pass;
   info[MOD_user_home]=cache_home;
   if(attr_dlocal[0])info[MOD_user_dlocal]=cache_dlocal;
   return 1;
  }
  if((ptr_aux1=strchr(email, '@'))){
   // Pegando dominio
   *ptr_aux1=0;
   local_strncpy(dominio, &ptr_aux1[1], sizeof(dominio));
   *ptr_aux1='@';
  }else dominio[0]=0;
  cp_str_var(linha, TAM_buffer, sch_user_filter, dominio, email);
  cp_str_var(buffer, TAM_buffer, sch_user_base, dominio, email);
  get_free_cnx();
  cont1=ldap_search_ext_s(ld[atual_cnx], buffer, LDAP_SCOPE_ONELEVEL, linha, attribs_user, 0 , NULL, NULL, NULL, LDAP_NO_LIMIT, &res);
  if(cont1 != LDAP_SUCCESS){
   snprintf(msg_error, TAM_msg_error, "LDAP Error: [%s]", ldap_err2string(cont1));
   free_cnx();
   return -1;
  }
  msg=ldap_first_message(ld[atual_cnx], res);
  if(msg){
   msg_type=ldap_msgtype(msg);
   if(msg_type == LDAP_RES_SEARCH_ENTRY){
    // Lendo atributos
    attr=ldap_first_attribute(ld[atual_cnx], msg, &ber);
    while(attr){
     // Pegando atributos
     if(((str_berval=ldap_get_values_len(ld[atual_cnx], msg, attr)) != NULL) && (str_berval[0])){
      if(!cmp_str(attr_mail, attr, tam_attr_mail))tam_user=local_strncpy(cache_user, str_berval[0]->bv_val, TAM_user);
      else if(!cmp_str(attr_pass, attr, tam_attr_pass))local_strncpy(cache_pass, str_berval[0]->bv_val, TAM_pass);
      else if(!cmp_str(attr_home, attr, tam_attr_home))local_strncpy(cache_home, str_berval[0]->bv_val, TAM_home);
      else if(!cmp_str(attr_maildir, attr, tam_attr_maildir))local_strncpy(cache_maildir, str_berval[0]->bv_val, TAM_maildir);
      else if(!cmp_str(attr_quota, attr, tam_attr_quota))cache_quota=atoi(str_berval[0]->bv_val);
      else if((attr_dlocal[0]) && (!cmp_str(attr_dlocal, attr, tam_attr_dlocal)))local_strncpy(cache_dlocal, str_berval[0]->bv_val, TAM_maildir);
     }
     ldap_memfree(attr);
     attr=ldap_next_attribute(ld[atual_cnx], msg, ber);
    } 
    info[MOD_user_user]=cache_user;
    info[MOD_user_pass]=cache_pass;
    info[MOD_user_home]=cache_home;
    if(attr_dlocal[0])info[MOD_user_dlocal]=cache_dlocal;
    ldap_msgfree(msg);
    free_cnx();
    return 1;         // --> Usuario existe
   }
  }
  free_cnx();
 }
 
 return 0;
}



//======================================================
// Retorna os alias                                     
// Retornos:                                            
// 0 = Nenhum alinas encontrado                         
// -1 = Erro                                            
// <numero de alias no buffer>                          
int mod_ret_alias(char *email, char *buffer_ret, int tam_buffer, int tam_item_buffer, unsigned int *prox_bloco){
 char dominio[200];
 char *ptr_aux1;
 
 // Apenas para testes
 //printf("LDAP: mod_alias_count(%s)\n", email);
 
 
 cont2=0;         // Vai guardar o numero de alias
 if(sch_alias_base[0]){
  if((ptr_aux1=strchr(email, '@'))){
   // Pegando dominio
   *ptr_aux1=0;
   local_strncpy(dominio, &ptr_aux1[1], sizeof(dominio));
   *ptr_aux1='@';
  }else dominio[0]=0;
  cp_str_var(linha, TAM_buffer, sch_alias_filter, dominio, email);
  cp_str_var(buffer, TAM_buffer, sch_alias_base, dominio, email);
  get_free_cnx();
  cont1=ldap_search_ext_s(ld[atual_cnx], buffer, LDAP_SCOPE_ONELEVEL, linha, attribs_alias, 0 , NULL, NULL, NULL, LDAP_NO_LIMIT, &res);
  if(cont1 != LDAP_SUCCESS){
   free_cnx();
   if(cont1 == LDAP_NO_SUCH_OBJECT)return 0;   // Alias nao existe
   snprintf(msg_error, TAM_msg_error, "LDAP Error: [%s]", ldap_err2string(cont1));
   return -1;
  }
  msg=ldap_first_message(ld[atual_cnx], res);
  cont3=0;        // Vai guardar o numero do item          
  while(msg && cont2 < tam_buffer){
   msg_type=ldap_msgtype(msg);
   if(msg_type == LDAP_RES_SEARCH_ENTRY){
    // Lendo atributos
    attr=ldap_first_attribute(ld[atual_cnx], msg, &ber);
    while(attr){
     // Procurando pelo atributo correto
     if(((str_berval=ldap_get_values_len(ld[atual_cnx], msg, attr)) != NULL) && (str_berval[0])){
      if((!cmp_str(attr_dstalias, attr, tam_attr_dstalias)) && (cont3++ >= *prox_bloco)){
       local_strncpy(buffer_ret, str_berval[0]->bv_val, tam_item_buffer);
       cont2++;
       buffer_ret+=tam_item_buffer;
       ldap_memfree(attr);
       goto Prox_msg;
      }
     }
     ldap_value_free_len(str_berval);
     ldap_memfree(attr);
     attr=ldap_next_attribute(ld[atual_cnx], msg, ber);
    }
   }
   Prox_msg:
   msg=ldap_next_message(ld[atual_cnx], msg);
  }
  ldap_msgfree(msg);
  free_cnx();
 }else msg=NULL;

 *prox_bloco+=cont2;
 return cont2; 
}




//======================================================
// Checa o usuario e senha passado                      
// Retornos:                                            
// 0 = Usuario nao encontrado                           
// 1 = Senha correta                                    
// 2 = Senha incorreta                                  
// -1 = Erro                                            
char mod_auth(char *email, char *senha, char *info[], int num_campos){
 char *ptr_aux1;
 char senha_md5_ldap[50];
 char senha_md5[16];
 // Apenas para testes
 //printf("LDAP: mod_auth(%s, %s)\n", email, senha);
 
 if(!(ret=mod_user_local(email, info, num_campos)))return 0; // Usurio nao existe
 else if(ret < 0)return ret;
 
 if(cache_pass[0] == '{'){
  // Metodo especificado
  if(!(ptr_aux1=strchr(cache_pass, '}'))){
   snprintf(msg_error, TAM_msg_error, "Invalid password format [%s]", email);
   return -1;
  }
  *ptr_aux1++=0;
  // -- CRYPT --
  if(!cmp_str(&cache_pass[1], "CRYPT\0", 6)){
   if(!strcmp((char *)crypt(senha, ptr_aux1), ptr_aux1))return 1;
  }else if(!cmp_str(&cache_pass[1], "MD5\0", 4)){
   if(strlen(ptr_aux1) > sizeof(senha_md5_ldap)){
    snprintf(msg_error, TAM_msg_error, "Password string is too big [%s]", email);
    return -1;
   }
   base64_decod(senha_md5_ldap, ptr_aux1);
   md5((uint8_t *)senha, strlen(senha), (uint8_t *)senha_md5);
   if(!cmp_str(senha_md5, senha_md5_ldap, 16))return 1;
  }else {
   snprintf(msg_error, TAM_msg_error, "Method used for password is not supported [%s]", email);
   return -1;
  } 
 }else if(!strcmp(senha, cache_pass))return 1;   // Sem metodo. Senha em texto puro
 
 return 2;
}

//======================================================
// Checa o usuario e senha passado                      
// Retornos:                                            
// 0 = Usuario nao encontrado                           
// 1 = Quota nao excedida                               
// 2 = Quota excedida                                   
// -1 = Erro                                            
char mod_test_quota(char *email, char *info[], int num_campos){
 char num_linhas=0;
 long long int ttotal=0;
 long long int ntotal=0;
 
 // Apenas para testes
 //printf("LDAP: mod_test_quota(%s)\n", email);
 if(!(ret=mod_user_local(email, info, num_campos)))return 0;
 else if(ret < 0)return ret;
 if(!cache_quota)return 1;
 snprintf(buffer, TAM_buffer, "%s/%s/maildirsize", cache_home, cache_maildir);
 arq_quota=open(buffer, O_RDWR | O_CREAT, 0666);
 if(arq_quota<0)return 1;        // Arquivo nao encontrado.
 cont2=0;
 do{
  b_lidos=read(arq_quota, buffer, TAM_buffer);
  for(cont1=0; cont1 < b_lidos; cont1++){
   if(buffer[cont1] == 10){
    linha[cont2]=0;
    if(!num_linhas)local_strncpy(linha_quota, linha, sizeof(linha_quota)-1);
    else {
     if(cont2 > 3){
      num_registros=sscanf(linha, "%lld %lld", &tam, &num);
      if(num_registros == 2){
       ttotal+=tam;
       ntotal+=num;
       if(ttotal < 0)ttotal=0;
       if(ntotal < 0)ntotal=0;
      }
     }
    }
    num_linhas++;
    cont2=0;
   }else if(buffer[cont1] != 13){
    linha[cont2]=buffer[cont1];
    cont2++;
   }
  }
 }while(b_lidos == TAM_buffer);
 if(num_linhas > 20 && !lockf(arq_quota, F_TLOCK, 0)){
  // Arquivo travado. Vamos resumi-lo
  b_lidos=sprintf(buffer, "%s\n%12lld %12lld\n", linha_quota, ttotal, ntotal);
  lseek(arq_quota, 0, 0);
  write(arq_quota, buffer, b_lidos);
  close(arq_quota);
  snprintf(buffer, TAM_buffer, "%s/%s/maildirsize", cache_home, cache_maildir);
  truncate(buffer, b_lidos);
 }else close(arq_quota);
 if(ttotal >= cache_quota)return 2;
 return 1;
}


//======================================================
// Finaiza o modulo  
char *mod_shutdown(void){
 
 return 0;
}

//======================================================
// Funcao para retornar um ponteiro contendo uma        
// mensagem de erro                                     
char *mod_msg_error(void){
 return &msg_error[0];
}
